(function (global, factory) {
  if (typeof define === "function" && define.amd) {
    define(['exports'], factory);
  } else if (typeof exports !== "undefined") {
    factory(exports);
  } else {
    var mod = {
      exports: {}
    };
    factory(mod.exports);
    global.signalTracker = mod.exports;
  }
})(this, function (exports) {
  'use strict';

  Object.defineProperty(exports, "__esModule", {
    value: true
  });

  function _classCallCheck(instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  }

  var _createClass = function () {
    function defineProperties(target, props) {
      for (var i = 0; i < props.length; i++) {
        var descriptor = props[i];
        descriptor.enumerable = descriptor.enumerable || false;
        descriptor.configurable = true;
        if ("value" in descriptor) descriptor.writable = true;
        Object.defineProperty(target, descriptor.key, descriptor);
      }
    }

    return function (Constructor, protoProps, staticProps) {
      if (protoProps) defineProperties(Constructor.prototype, protoProps);
      if (staticProps) defineProperties(Constructor, staticProps);
      return Constructor;
    };
  }();

  const Signals = imports.signals;

  /**
   * Tracks objects implementing the GJS signal emitter interface.
   *
   * This class should be used on a per-extension basis. Call {@link clean} when
   * it is disabled.
   *
   * @see {@link connect}
   * @see {@link imports.signals}
   */

  let SignalTracker = function () {
    function SignalTracker() {
      _classCallCheck(this, SignalTracker);

      Signals.addSignalMethods(this.cleaner = {});
    }

    /**
     * Connects an object's signal to a handler and tracks it until {@link clean}
     * is called.
     *
     * @param {Object} obj the object to track.
     * @param {String} sig the object's signal to track.
     * @param {Function} cb the signal handler.
     * @return {Function} the callback to manually stop the tracking.
     */


    _createClass(SignalTracker, [{
      key: 'connect',
      value: function connect(obj, sig, cb) {
        const oid = obj.connect(sig, cb);
        const cid = this.cleaner.connect(null, () => obj.disconnect(oid));
        return () => {
          this.cleaner.disconnect(cid);obj.disconnect(oid);
        };
      }
    }, {
      key: 'clean',
      value: function clean() {
        this.cleaner.emit();
        this.cleaner.disconnectAll();
      }
    }]);

    return SignalTracker;
  }();

  exports.default = SignalTracker;

  let ShellWindowTracker = exports.ShellWindowTracker = function () {

    /**
     * @param {SignalTracker} tracker a signal tracker.
     */
    function ShellWindowTracker(tracker) {
      _classCallCheck(this, ShellWindowTracker);

      this.tracker = tracker;
    }

    /**
     * Executes a callback on all current windows, and also whenever a new one is
     * created.
     *
     * @param {Function(gi.Meta.Window)} cb the callback.
     * @return {Function} the callback to manually stop the tracking.
     */


    _createClass(ShellWindowTracker, [{
      key: 'onEvery',
      value: function onEvery(cb) {
        for (let a of global.get_window_actors()) cb(a.meta_window);
        return this.tracker.connect(global.display, 'window-created', (_, win) => cb(win));
      }

      /**
       * Connects a window's signal to a handler and tracks it with the underlying
       * signal tracker.
       *
       * @param {gi.Meta.Window} win the window to track.
       * @param {String} sig the window's signal to track.
       * @param {Function} cb the signal handler.
       * @return {Function} the callback to manually stop the tracking.
       */

    }, {
      key: 'connect',
      value: function connect(win, sig, cb) {
        const dc = this.tracker.connect(win, sig, cb);
        return this.tracker.connect(win, 'unmanaged', dc);
      }

      /**
       * Connects every window's signal to a handler and tracks it with the
       * underlying signal tracker.
       *
       * @param {String} sig the window's signal to track.
       * @param {Function} cb the signal handler.
       * @return {Function} the callback to manually stop the tracking.
       * @see {@link connect}
       * @see {@link onEvery}
       */

    }, {
      key: 'connectEvery',
      value: function connectEvery(sig, cb) {
        return this.onEvery(win => this.connect(win, sig, cb));
      }
    }]);

    return ShellWindowTracker;
  }();

  let ShellWorkspaceTracker = exports.ShellWorkspaceTracker = function () {

    /**
     * @param {SignalTracker} tracker a signal tracker.
     */
    function ShellWorkspaceTracker(tracker) {
      _classCallCheck(this, ShellWorkspaceTracker);

      this.tracker = tracker;
    }

    /**
     * Executes a callback on all current workspaces, and also whenever a new one
     * is created.
     *
     * @param {Function(gi.Meta.Workspace)} cb the callback.
     * @return {Function} the callback to manually stop the tracking.
     */


    _createClass(ShellWorkspaceTracker, [{
      key: 'onEvery',
      value: function onEvery(cb) {
        for (let i = 0; i < global.screen.n_workspaces; i++) cb(global.screen.get_workspace_by_index(i));
        return this.tracker.connect(global.screen, 'workspace-added', (_, i) => cb(global.screen.get_workspace_by_index(i)));
      }

      /**
       * Connects a workspace's signal to a handler and tracks it with the
       * underlying signal tracker.
       *
       * @param {gi.Meta.Workspace} win the workspace to track.
       * @param {String} sig the workspace's signal to track.
       * @param {Function} cb the signal handler.
       * @return {Function} the callback to manually stop the tracking.
       */

    }, {
      key: 'connect',
      value: function connect(ws, sig, cb) {
        const dc = this.tracker.connect(ws, sig, cb);
        return this.tracker.connect(global.screen, 'workspace-removed', dc);
      }

      /**
       * Connects every workspace's signal to a handler and tracks it with the
       * underlying signal tracker.
       *
       * @param {String} sig the workspace's signal to track.
       * @param {Function} cb the signal handler.
       * @return {Function} the callback to manually stop the tracking.
       * @see {@link connect}
       * @see {@link onEvery}
       */

    }, {
      key: 'connectEvery',
      value: function connectEvery(sig, cb) {
        return this.onEvery(ws => this.connect(ws, sig, cb));
      }
    }]);

    return ShellWorkspaceTracker;
  }();
});
