const Signals = imports.signals;

/**
 * Tracks objects implementing the GJS signal emitter interface.
 *
 * This class should be used on a per-extension basis. Call {@link clean} when
 * it is disabled.
 *
 * @see {@link connect}
 * @see {@link imports.signals}
 */
export default class SignalTracker {

  constructor() {
    Signals.addSignalMethods(this.cleaner = {});
  }

  /**
   * Connects an object's signal to a handler and tracks it until {@link clean}
   * is called.
   *
   * @param {Object} obj the object to track.
   * @param {String} sig the object's signal to track.
   * @param {Function} cb the signal handler.
   * @return {Function} the callback to manually stop the tracking.
   */
  connect(obj, sig, cb) {
    const oid = obj.connect(sig, cb);
    const cid = this.cleaner.connect(null, () => obj.disconnect(oid));
    return () => { this.cleaner.disconnect(cid); obj.disconnect(oid); };
  }

  /**
   * Disconnects and untracks all tracked signals.
   */
  clean() {
    this.cleaner.emit();
    this.cleaner.disconnectAll();
  }
}

/**
 * Tracks GNOME Shell's windows.
 *
 * The underlying signal tracker is not cleaned automatically.
 *
 * @see {@link onEvery}
 * @see {@link connectEvery}
 */
export class ShellWindowTracker {

  /**
   * @param {SignalTracker} tracker a signal tracker.
   */
  constructor(tracker) {
    this.tracker = tracker;
  }

  /**
   * Executes a callback on all current windows, and also whenever a new one is
   * created.
   *
   * @param {Function(gi.Meta.Window)} cb the callback.
   * @return {Function} the callback to manually stop the tracking.
   */
  onEvery(cb) {
    for (let a of global.get_window_actors())
      cb(a.meta_window);
    return this.tracker.connect(global.display, 'window-created',
        (_, win) => cb(win));
  }

  /**
   * Connects a window's signal to a handler and tracks it with the underlying
   * signal tracker.
   *
   * @param {gi.Meta.Window} win the window to track.
   * @param {String} sig the window's signal to track.
   * @param {Function} cb the signal handler.
   * @return {Function} the callback to manually stop the tracking.
   */
  connect(win, sig, cb) {
    const dc = this.tracker.connect(win, sig, cb);
    return this.tracker.connect(win, 'unmanaged', dc);
  }

  /**
   * Connects every window's signal to a handler and tracks it with the
   * underlying signal tracker.
   *
   * @param {String} sig the window's signal to track.
   * @param {Function} cb the signal handler.
   * @return {Function} the callback to manually stop the tracking.
   * @see {@link connect}
   * @see {@link onEvery}
   */
  connectEvery(sig, cb) {
    return this.onEvery(win => this.connect(win, sig, cb));
  }
}

/**
 * Tracks GNOME Shell's workspaces.
 *
 * The underlying signal tracker is not cleaned automatically.
 *
 * @see {@link onEvery}
 * @see {@link connectEvery}
 */
export class ShellWorkspaceTracker {

  /**
   * @param {SignalTracker} tracker a signal tracker.
   */
  constructor(tracker) {
    this.tracker = tracker;
  }

  /**
   * Executes a callback on all current workspaces, and also whenever a new one
   * is created.
   *
   * @param {Function(gi.Meta.Workspace)} cb the callback.
   * @return {Function} the callback to manually stop the tracking.
   */
  onEvery(cb) {
    for (let i = 0; i < global.screen.n_workspaces; i++)
      cb(global.screen.get_workspace_by_index(i));
    return this.tracker.connect(global.screen, 'workspace-added',
        (_, i) => cb(global.screen.get_workspace_by_index(i)));
  }

  /**
   * Connects a workspace's signal to a handler and tracks it with the
   * underlying signal tracker.
   *
   * @param {gi.Meta.Workspace} win the workspace to track.
   * @param {String} sig the workspace's signal to track.
   * @param {Function} cb the signal handler.
   * @return {Function} the callback to manually stop the tracking.
   */
  connect(ws, sig, cb) {
    const dc = this.tracker.connect(ws, sig, cb);
    return this.tracker.connect(global.screen, 'workspace-removed', dc);
  }

  /**
   * Connects every workspace's signal to a handler and tracks it with the
   * underlying signal tracker.
   *
   * @param {String} sig the workspace's signal to track.
   * @param {Function} cb the signal handler.
   * @return {Function} the callback to manually stop the tracking.
   * @see {@link connect}
   * @see {@link onEvery}
   */
  connectEvery(sig, cb) {
    return this.onEvery(ws => this.connect(ws, sig, cb));
  }
}
